from datetime import datetime, date, timedelta, time
from decimal import Decimal
from json import JSONEncoder, dumps, loads
from base64 import b64encode
from pydicts import casts, exceptions

# Forma en que debe parsearse los Decimals
class DecimalsWay:
    DecimalString=1  #Uses a String with decimal to detect decimals "Decimal('12.122')"
    String=2  #"12.122"
    Float=3  # 12.122

## Usa
class MyJSONEncoder(JSONEncoder):
    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime):
            if casts.is_aware(o):
                return casts.dtaware2str(o)
            else: #naive
                return casts.dtnaive2str(o)
        elif isinstance(o, date):
            return o.isoformat()
        elif isinstance(o, time):
            if o.utcoffset() is not None: #If it's aware
                raise ValueError("JSON can't represent timezone-aware times.")
            return casts.time2str(o)
        elif isinstance(o, timedelta):
            return casts.timedelta2str(o)
        elif isinstance(o, Decimal):
            return f"Decimal('{o}')"
        elif o.__class__.__name__ in ("Promise", "__proxy__"): #django.utils.functional
            return str(o)
        elif isinstance(o, bytes):
            return b64encode(o).decode("UTF-8")
        elif o.__class__.__name__=="Percentage":
            return o.value
        elif o.__class__.__name__=="Currency":
            return o.amount
        else:
            return super().default(o)

## Usa decimals como floats normalmwente para JS
class MyJSONEncoderDecimalsAsString(MyJSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return str(o)
        else:
            return super().default(o)
            
## Usa decimals como floats normalmwente para JS
class MyJSONEncoderDecimalsAsFloat(MyJSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return float(o)
        else:
            return super().default(o)

                
def MyJSONEncoder_dumps(r, indent=4):
    return dumps(r, cls=MyJSONEncoder, indent=indent)
                
def MyJSONEncoderDecimalsAsString_dumps(r, indent=4):
    return dumps(r, cls=MyJSONEncoderDecimalsAsString, indent=indent)
                
def MyJSONEncoderDecimalsAsFloat_dumps(r, indent=4):
    return dumps(r, cls=MyJSONEncoderDecimalsAsFloat, indent=indent)


def MyJSONEncoder_loads(s):
    def hooks_MyJSONEncoder(iter_value):
        return hooks(iter_value, DecimalsWay.DecimalString)
    ##############################
    return loads(s, object_hook=hooks_MyJSONEncoder)

def MyJSONEncoderDecimalsAsFloat_loads(s):
    
    def hooks_MyJSONEncoderAsFloat(iter_value):
        return hooks(iter_value, DecimalsWay.Float)
    ####################################
    return loads(s, object_hook=hooks_MyJSONEncoderAsFloat)        

def MyJSONEncoderDecimalsAsString_loads(s):
    def hooks_MyJSONEncoderAsString(iter_value):
        return hooks(iter_value, DecimalsWay.String)
    ######################################
    return loads(s, object_hook=hooks_MyJSONEncoderAsString)
    
    
def hooks(iter_value, decimals_way):
    """
        Iterates a dict or list to cast decimals and dtaware in json.loads using objeck_hook
    """
    def guess_cast(o, decimal_way):
        if decimal_way==DecimalsWay.DecimalString:
            if o.__class__==str and o.startswith("Decimal("):
                try:
                    return eval(o)
                except:
                    pass
                    
        # Guess date
        try:
            return casts.str2date(o)
        except exceptions.CastException:
            pass
            
        #Guess dtaware
        try:
            return casts.str2dtaware(o,"JsUtcIso")
        except exceptions.CastException:
            pass

        #Guess dtnaive
        try:
            return casts.str2dtnaive(o,"JsIso")
        except exceptions.CastException:
            pass

        #Guess time
        try:
            return casts.str2time(o)
        except exceptions.CastException:
            pass
                    
        # Guess date
        try:
            return casts.str2timedelta(o)
        except exceptions.CastException:
            pass

        #Guess Bytes
        try:
            b64bytes=casts.str2bytes(o)# o is a b64 string
            return casts.base64bytes2bytes(b64bytes)
        except:
            pass

        return o
    ########################################################
    
    if isinstance(iter_value, dict):
        for k, v in iter_value.items():
            if isinstance(v, dict):
                iter_value[k]=hooks(v, decimals_way)
            elif isinstance(iter_value, list):
                for i in v:
                    i=hooks(v, decimals_way)
            else:
                guessed=guess_cast(v, decimals_way)
#                print("GUESS_CAST", iter_value[k], decimals_way, "GOT", guessed, guessed.__class__)
                iter_value[k]=guessed
    elif isinstance(iter_value, list):
        for i in v:
            i=hooks(i, decimals_way)
    return iter_value
    
    




from datetime import datetime, date
from decimal import Decimal
from pydicts import dod
from pytest import fixture

empty_lod=[]

@fixture(autouse=True)
def reload_lod_():
    d={"a": datetime.now(), "b": date.today(), "c": Decimal('12.32'), "d": None, "e": int(12), "f":None, "g":True, "h":False,  "nested": {"x":1, "y":2, "z":{"n":4, "m":5, "o":6}}}
    global dod_
    dod_={}
    dod_["first"]=d
    dod_["second"]=d


def test_dod_print():
    dod.dod_print(dod_)



from decimal import Decimal
from pydicts import lod, exceptions
from importlib.resources import files



def currencies_list():
    with open(files("pydicts") / 'currencies.json') as f:
        lod_=eval(f.read())
    return lod_

def currencies_odod():
    return lod.lod2odod(currencies_list(), "code")
    
currencies=currencies_odod()

## Class to manage currencies in officegenerator
##
## The symbol is defined by code with self.symbol()
class Currency:
    def __init__(self, amount=None,  currency="EUR") :
        if amount==None:
            self.amount=Decimal(0)
        else:
            self.amount=Decimal(str(amount))
        self.currency=currency

    def __eq__(self, other):
        if self.amount==other.amount and self.currency==other.currency:
            return True
        return False

    def __add__(self, other):
        """Si las divisas son distintas, queda el resultado con la divisa del primero"""
        if self.currency==other.currency:
            return Currency(self.amount+other.amount, self.currency)
        else:
            raise exceptions.CurrencyOperationsException("Before adding, please convert to the same currency")

    def __sub__(self, other):
        """Si las divisas son distintas, queda el resultado con la divisa del primero"""
        if self.currency==other.currency:
            return Currency(self.amount-other.amount, self.currency)
        else:
            raise exceptions.CurrencyOperationsException("Before substracting, please convert to the same currency")

    def __lt__(self, other):
        if self.currency==other.currency:
            if self.amount < other.amount:
                return True
            return False
        else:
            raise exceptions.CurrencyOperationsException("Before lt ordering, please convert to the same currency")
    
    def __le__(self, other):
        if self.currency==other.currency:
            if self.amount <= other.amount:
                return True
            return False
        else:
            raise exceptions.CurrencyOperationsException("Before lt ordering, please convert to the same currency")
            
    def __gt__(self, other):
        if self.currency==other.currency:
            if self.amount > other.amount:
                return True
            return False
        else:
            raise exceptions.CurrencyOperationsException("Before lt ordering, please convert to the same currency")
            
    def __ge__(self, other):
        if self.currency==other.currency:
            if self.amount >= other.amount:
                return True
            return False
        else:
            raise exceptions.CurrencyOperationsException("Before lt ordering, please convert to the same currency")

    ## Si las divisas son distintas, queda el resultado con la divisa del primero
    ##
    ## En caso de querer multiplicar por un numero debe ser despues. For example: other*4
    def __mul__(self, other):
        if other.__class__.__name__ in ("int",  "float", "Decimal"):
            return Currency(self.amount*other, self.currency)

        if self.currency==other.currency:
            print(self.amount, other.amount, self.amount*other.amount)
            return Currency(self.amount*other.amount, self.currency)
        else:
            raise exceptions.CurrencyOperationsException("Before multiplying, please convert to the same currency")

    def __truediv__(self, other):
        """Si las divisas son distintas, queda el resultado con la divisa del primero"""
        if self.currency==other.currency:
            return Currency(self.amount/other.amount, self.currency)
        else:
            raise exceptions.CurrencyOperationsException("Before true dividing, please convert to the same currency")

    def __repr__(self):
        return self.string(2)

    ## Returs a typical currency string
    ## @param digits int that defines the number of decimals. 2 by default
    ## @return string
    def string(self,   digits=2):
        return "{} {}".format(round(self.amount, digits), currencies[self.currency]["symbol"])

    def isZero(self):
        if self.amount==Decimal(0):
            return True
        else:
            return False

    def isGETZero(self):
        if self.amount>=Decimal(0):
            return True
        else:
            return False

    def isGTZero(self):
        if self.amount>Decimal(0):
            return True
        else:
            return False

    def isLTZero(self):
        if self.amount<Decimal(0):
            return True
        else:
            return False

    def isLETZero(self):
        if self.amount<=Decimal(0):
            return True
        else:
            return False

    def __neg__(self):
        """Devuelve otro other con el amount con signo cambiado"""
        return Currency(-self.amount, self.currency)

    def round(self, digits=2):
        return round(self.amount, digits)








from decimal import Decimal

## Class to manage percentages in spreadsheets
## Converts values to decimal (Value is a decimal)
class Percentage:
    def __init__(self, numerator=None, denominator=None):
        
        self.setValue(numerator, denominator)

    def __repr__(self):
        return self.string()

    def __neg__(self):
        if self.value is None:
            return Percentage(None)
        return Percentage(-self.value, 1)

    def __lt__(self, other):
        if self.value is None or other.value is None:
            return False
        
        
        if self.value<other.value:
            return True
        return False

    def __eq__(self, b):
        
        return self.value==b.value

    def __add__(self,p):
        if self.value is None or p.value is None:
            return None
        
        return self.__class__(self.value+p.value,1)

    def __sub__(self, p):
        if self.value is None or p.value is None:
            return None
        
        return self.__class__(self.value-p.value,1)

    def __mul__(self, value):
        try:
            if value.__class__==Percentage:
                return Percentage(self.value*value.value, 1)
        except:
            return Percentage(None)

    def __truediv__(self, other):
        try:
            r=self.value/other.value
        except:
            r=None
        return Percentage(r, 1)

    def setValue(self, numerator,  denominator):
        try:
            self.value=Decimal(numerator)/Decimal(denominator)
        except:
            self.value=None

    def value_100(self):
        if self.value==None:
            return None
        else:
            return self.value*Decimal(100)

    ## @return percentage float value
    def float_100(self):
        if self.value is None:
            return None
        return float(self.value_100())

    ## @return percentage float value
    def float(self):
        if self.value is None:
            return None
        return float(self.value)

    def string(self, rnd=2):
        if self.value is None:
            return "None %"
        return "{} %".format(round(self.value_100(), rnd))

    ## Returns if the percentage is valid. I mean it's value different of None
    def isValid(self):
        if self.value!=None:
            return True
        return False

    def isGETZero(self):
        if self.isValid() and self.value>=0:
            return True
        return False

    def isGTZero(self):
        if self.isValid() and self.value>0:
            return True
        return False

    def isLTZero(self):
        if self.isValid() and self.value<0:
            return True
        return False
        
    def isLETZero(self):
        if self.isValid() and self.value<=0:
            return True
        return False

## Calculates porcentage to pass from a to b
## @param a. Can be an object divisible and that can be substracted
def percentage_between(a,b):
    try:
        return Percentage(b-a,a)
    except:
        return Percentage()



"""
    All cast will raise a CastException(message) if cast fails
    All cast will have a parameter ignore_exception=False, to ignore exceptions and return ignore_exception_value
    All cast hava a method ignore_exception_value=None
"""

from decimal import Decimal
from datetime import timedelta, date, datetime, time
from gettext import translation
from importlib.resources import files
from pydicts import exceptions
from re import match
from zoneinfo import ZoneInfo
from base64 import b64encode, b64decode
from isodate import parse_duration, duration_isoformat


try:
    t=translation('pydicts', files("pydicts") / 'locale')
    _=t.gettext
except:
    _=str
    
def object_or_empty(v):
    """
        Returns and empty string if None, else return value
    """
    return "" if v is None else v

def str2decimal(value, ignore_exception=False, ignore_exception_value=None):
    """
        Converts a string  to a decimal
        Parameters:
            - decimal_separator. Symbol to separate decimals. For example 12.121212 (decimal_separator=".") 12.122,1223 (decimal_separator=",")
    """
    original=value
    error=f"Error in Pydicts.cast.str2decimal method. Value: {original} Value class: {value.__class__.__name__}"

    if is_noe(value) or not value.__class__ ==str:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        return Decimal(value)
    except:

        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

def str2bool(value, ignore_exception=False, ignore_exception_value=None):
    """
        Converts strings True or False to boolean
        @param value String
        @return Boolean
    """
    original=value
    error=f"Error in Pydicts.cast.str2bool method. Value: {original} Value class: {value.__class__.__name__}"

    if is_noe(value) or not value.__class__ ==str:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value
        
    if value=="0" or value.lower()=="false":
        return False
    elif value=="1" or value.lower()=="true":
        return True
    else:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

def none2alternative(value, alternative):
    """
        If a value is None, returns an alternative
    """
    if value is None:
        return alternative
    return value

def bytes2str(value, code='UTF-8', ignore_exception=False, ignore_exception_value=None):
    """
        Bytes 2 string
    """ 
    
    original=value
    error=f"Error in Pydicts.cast.bytes2str method. Value: {original} Value class: {value.__class__.__name__}"

    if value is None or not value.__class__==bytes:

        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        return value.decode(code)
    except:

        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    
def str2bytes(value, code='UTF8', ignore_exception=False, ignore_exception_value=None):
    """
        String 2 bytes
    """       
    original=value
    error=f"Error in Pydicts.cast.str2bytes method. Value: {original} Value class: {value.__class__.__name__}"

    if value is None or not value.__class__==str:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        return value.encode(code)
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value
            
        
def base64bytes2bytes(value, ignore_exception=False, ignore_exception_value=None):
    """
        Bytes 2 string
    """ 
    
    original=value
    error=f"Error in Pydicts.cast.base642bytes method. Value: {original} Value class: {value.__class__.__name__}"

    if value is None or not value.__class__==bytes:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        return b64decode(value)
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    
def bytes2base64bytes(value, ignore_exception=False, ignore_exception_value=None):
    """
        String 2 bytes
    """       
    original=value
    error=f"Error in Pydicts.cast.bytes2base64 method. Value: {original} Value class: {value.__class__.__name__}"

    if value is None or not value.__class__==bytes:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        return b64encode(value)
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value    

### Returns if a datetime is aware
def is_aware(dt):
    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        return False
    return True

## Returns if a datetime is naive
def is_naive(dt):
    return not is_aware(dt)

def is_noe(value):
    """
        Returns a boolean. True if value is None or an empty string.
    """
    if value is None:
        return True
    if value =="":
        return True
    return False

## Function to create a datetime aware object
## @param date datetime.date object
## @param hour hour object
## @param tz_name String with datetime tz_name name. For example "Europe/Madrid"
## @return datetime aware
def dtaware(date_, time_, tz_name):
    dt_naive=dtnaive(date_, time_)
    return dtnaive2dtaware(dt_naive, tz_name)

def dtnaive2dtaware(dtnaive, tz_name):
    return dtnaive.replace(tzinfo=ZoneInfo(tz_name))

def dtaware2dtnaive(dtaware):
    if not is_aware(dtaware): 
        raise exceptions.CastException(f"{dtaware} should be a dtaware")
    return dtaware.replace(tzinfo=None)

def dtaware_now(tzname=None):
    """
        If tzname is None: returns UTC dtaware
    """
    utc_aware=dtnaive2dtaware(dtnaive_now(), 'UTC')
    if tzname is None:
        return utc_aware
    else:
        return dtaware_changes_tz(utc_aware, tzname)

def dtnaive_now():
    return datetime.now()

## Function to create a datetime aware object
## @param date datetime.date object
## @param hour hour object
## @return datetime naive
def dtnaive(date_, hour):
    return datetime(date_.year,  date_.month,  date_.day,  hour.hour,  hour.minute,  hour.second, hour.microsecond)


## Returns a date with the first date of the month
## @param year Year to search fist day
## @param month Month to search first day
def date_first_of_the_month(year, month):
    return date(year, month, 1)

## Returns a date with the last date of the month
## @param year Year to search last day
## @param month Month to search last day
def date_last_of_the_month(year, month):
    if month==12:
        return date(year, month, 31)
    return date(year, month+1, 1)-timedelta(days=1)

## Returns a date with the first date of the year
## @param year Year to search first day
def date_first_of_the_year(year):
    return date_first_of_the_month(year, 1)

## Returns a date with the last date of the year
## @param year Year to search last day
def date_last_of_the_year(year):
    return date_last_of_the_month(year,12)

## Returns a date with the first date of the month after x months
## @param year Year to search  day
## @param month Month to search day
## @param x Number of months after parameters. Cab be positive to add months or negative to substract months
def date_first_of_the_next_x_months(year, month, x):
    if x>=0:
        first=date(year, month, 1)
        for i in range(x):
            last=date_last_of_the_month(first.year, first.month)
            first=last+timedelta(days=1)
        return first
    else:#<0
        first=date(year, month, 1)
        for i in range(abs(x)):
            last_before=first-timedelta(days=1)
            first=date_first_of_the_month(last_before.year, last_before.month)
        return first    

## Returns a date with the last date of the month after x months
## @param year Year to search  day
## @param month Month to search day
## @param x Number of months after parameters. Cab be positive to add months or negative to substract months
def date_last_of_the_next_x_months(year, month, x):
    first=date_first_of_the_next_x_months(year, month, x)
    return date_last_of_the_month(first.year, first.month)

def dtaware_month_end(year, month, tz_name):
    return dtaware_day_end_from_date(date_last_of_the_month(year, month), tz_name)
    
## Returns an aware datetime with the start of year
def dtaware_year_start(year, tz_name):
    return dtaware_day_start_from_date(date(year, 1, 1), tz_name)
    
## Returns an aware datetime with the last of year
def dtaware_year_end(year, tz_name):
    return dtaware_day_end_from_date(date(year, 12, 31), tz_name)
    
def dtaware_day_end(dt, tz_name):
    """
        Returns the last  datetime (microsecond  level) of the  day in tz_name zone
    """
    if is_naive():
        raise exceptions.CastException(_("A datetime with timezone is needed"))
    dt=dtaware_changes_tz(dt, tz_name)
    return dt.replace(hour=23, minute=59, second=59, microsecond=999999)
    
def dtnaive_day_end(dt):
    """
        Returns the last  datetime (microsecond  level) of the  day in naive format
    """
    if is_aware(dt):
        raise exceptions.CastException(_("A datetime without timezone is needed"))
    return dt.replace(hour=23, minute=59, second=59, microsecond=999999)

## Returns the end of the day dtnaive from a date
def dtnaive_day_end_from_date(date_):
    dt=datetime(date_.year, date_.month, date_.day)
    return dtnaive_day_end(dt)

## Returns the end of the day dtaware in utc format
def dtaware_day_end_from_date(date, tz_name):
    return dtaware(date, time(23, 59, 59, 999999), tz_name)
    
## Returns a dtnaive or dtawre (as parameter) with the end of the day
def dtnaive_day_start(dt):
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)

## Returns a dtnaive or dtawre (as parameter) with the end of the day in zone tz_name
def dtaware_day_start(dt, tz_name):
    if is_naive():
        raise exceptions.CastException(_("A datetime with timezone is needed"))
    dt=dtaware_changes_tz(dt, tz_name)
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)

## Returns the end of the day dtnaive from a date
def dtnaive_day_start_from_date(date_):
    dt=datetime(date_.year, date_.month, date_.day)
    return dtnaive_day_start(dt)

## Returns the end of the day dtaware of the tz_name timezone from a date
def dtaware_day_start_from_date(date, tz_name):
    return dtaware(date, time(0, 0, 0, 0), tz_name)

## Returns the start of a month in utc format
def dtaware_month_start(year, month, tz_name):
    return dtaware_day_start_from_date(date(year, month, 1), tz_name)
#    
#def month2int(s):
#    """
#        Converts a month string to a int
#    """
#    if s in ["Jan", "Ene", "Enero", "January", "enero", "january"]:
#        return 1
#    if s in ["Feb", "Febrero", "February", "febrero", "february"]:
#        return 2
#    if s in ["Mar", "Marzo", "March", "marzo", "march"]:
#        return 3
#    if s in ["Apr", "Abr", "April", "Abril", "abril", "april"]:
#        return 4
#    if s in ["May", "Mayo", "mayo", "may"]:
#        return 5
#    if s in ["Jun", "June", "Junio", "junio", "june"]:
#        return 6
#    if s in ["Jul", "July", "Julio", "julio", "july"]:
#        return 7
#    if s in ["Aug", "Ago", "August", "Agosto", "agosto", "august"]:
#        return 8
#    if s in ["Sep", "Septiembre", "September", "septiembre", "september"]:
#        return 9
#    if s in ["Oct", "October", "Octubre", "octubre", "october"]:
#        return 10
#    if s in ["Nov", "Noviembre", "November", "noviembre", "november"]:
#        return 11
#    if s in ["Dic", "Dec", "Diciembre", "December", "diciembre", "december"]:
#        return 12


def str2time(value, format="JsIso", ignore_exception=False, ignore_exception_value=None):
    original=value
    allowed=["HH:MM", "HH:MM:SS","JsIso","HH:MMxx"]
    error=f"Error in Pydicts.cast.str2time method. Value: {original} Value class: {value.__class__.__name__} Format: {format} Allowed: {allowed}"
    if format not in  allowed or value.__class__!=str:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        if format=="HH:MM":#12:12
            a=value.split(":")
            return time(int(a[0]), int(a[1]))
        elif format=="HH:MM:SS":#12:12:12
            a=value.split(":")
            return time(int(a[0]), int(a[1]), int(a[2]))
        elif format=="HH:MMxx": #5:12am o pm
            value=value.upper()
            value=value.replace("AM", "")
            if value.find("PM"):
                value=value.replace("PM", "")
                points=value.split(":")
                return time(int(points[0])+12, int(points[1]))
            else:#AM
                points=value.split(":")
                return time(int(points[0]), int(points[1]))
        elif format=="JsIso":#23:00:00.000000  
            if not ":" in value:
                raise exceptions.CastException(error)
            return time.fromisoformat(value)
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

## Converts a time to a string
def time2str(value, format="JsIso" , ignore_exception=False, ignore_exception_value=None):
    original=value
    allowed=["HH:MM", "HH:MM:SS","Xulpymoney", "JsIso"]
    error=f"Error in Pydicts.cast.str2time method. Value: {original} Value class: {value.__class__.__name__} Format: {format} Allowed: {allowed}"
    if format not in  allowed or value.__class__!=time:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        if format=="Xulpymoney":
            if value.microsecond in (4, 5):
                return str(value)[11:-13]
            else:
                return str(value)[11:-6]
        elif format=="HH:MM":
            return ("{}:{}".format(str(value.hour).zfill(2), str(value.minute).zfill(2)))
        elif format=="HH:MM:SS":
            return ("{}:{}:{}".format(str(value.hour).zfill(2), str(value.minute).zfill(2), str(value.second).zfill(2)))
        elif format=="JsIso":
            return(str(value))
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value


def str2date(value, format="YYYY-MM-DD", ignore_exception=False, ignore_exception_value=None):
    original=value
    allowed=["YYYY-MM-DD", "DD/MM/YYYY", "DD.MM.YYYY", "DD/MM", "JsIso"]
    error=f"Error in Pydicts.cast.str2date method. Value: {original} Value class: {value.__class__.__name__} Format: {format} Allowed: {allowed}"
    if format not in  allowed or value.__class__!=str:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        if format in ["YYYY-MM-DD", "JsIso"]:
            d=value.split("-")
            return date(int(d[0]), int(d[1]),  int(d[2]))
        if format=="DD/MM/YYYY": #DD/MM/YYYY
            d=value.split("/")
            return date(int(d[2]), int(d[1]),  int(d[0]))
        if format=="DD.MM.YYYY": #DD.MM.YYYY
            d=value.split(".")
            return date(int(d[2]), int(d[1]),  int(d[0]))
        if format=="DD/MM": #DD/MM
            d=value.split("/")
            return date(date.today().year, int(d[1]),  int(d[0]))
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

def str2dtnaive(value, format="JsIso", ignore_exception=False, ignore_exception_value=None):
    original=value
    allowed=["%Y%m%d%H%M","%Y-%m-%d %H:%M:%S","%d/%m/%Y %H:%M","%d %m %H:%M %Y","%Y-%m-%d %H:%M:%S.","%H:%M:%S", '%b %d %H:%M:%S', "JsIso"]
    error=f"Error in Pydicts.cast.str2dtnaive method. Value: {original} Value class: {value.__class__.__name__} Format: {format} Allowed: {allowed}"
    if format not in  allowed or value.__class__!=str:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        if format=="%Y%m%d%H%M":
            dat=datetime.strptime( value, format )
            return dat
        if format=="%Y-%m-%d %H:%M:%S":#2017-11-20 23:00:00
            return datetime.strptime( value, format )
        if format=="%d/%m/%Y %H:%M":#20/11/2017 23:00
            return datetime.strptime( value, format )
        if format=="%d %m %H:%M %Y":#27 1 16:54 2017. 1 es el mes convertido con month2int
            return datetime.strptime( value, format)
        if format=="%Y-%m-%d %H:%M:%S.":#2017-11-20 23:00:00.000000  ==>  microsecond. Notice the point in format
            arrPunto=value.split(".")
            micro=int(arrPunto[1]) if len(arrPunto)==2 else 0
            dt=datetime.strptime( arrPunto[0],  "%Y-%m-%d %H:%M:%S" )
            dt=dt+timedelta(microseconds=micro)
            return dt
        if format=="%H:%M:%S": 
            tod=date.today()
            a=value.split(":")
            return datetime(tod.year, tod.month, tod.day, int(a[0]), int(a[1]), int(a[2]))
        if format=='%b %d %H:%M:%S': #Apr 26 07:50:44. Year is missing so I set to current
            s=f"{date.today().year} {value}"
            return datetime.strptime(s, '%Y %b %d %H:%M:%S')
        if format=="JsIso": #2021-08-21T06:27:38.294
            if "Z" in value:
                raise exceptions.CastException(error)
            value=value.replace("T"," ")
            dtnaive=str2dtnaive(value,"%Y-%m-%d %H:%M:%S.")

            return dtnaive
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

def str2dtaware(value, format="JsUtcIso", tz_name='UTC', ignore_exception=False, ignore_exception_value=None):
    original=value
    allowed=["%Y-%m-%d %H:%M:%S%z","%Y-%m-%d %H:%M:%S.%z", "JsUtcIso"]
    error=f"Error in Pydicts.cast.str2dtaware method. Value: {original} Value class: {value.__class__.__name__} Format: {format} Allowed: {allowed}"
    if format not in  allowed or value.__class__!=str:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value
    try:
        if format=="%Y-%m-%d %H:%M:%S%z":#2017-11-20 23:00:00+00:00
            s=value[:-3]+value[-2:]
            dt=datetime.strptime( s, format )
            return dtaware_changes_tz(dt, tz_name)
        if format=="%Y-%m-%d %H:%M:%S.%z":#2017-11-20 23:00:00.000000+00:00  ==>  microsecond. Notice the point in format
            s=value[:-3]+value[-2:]#quita el :
            arrPunto=s.split(".")
            s=arrPunto[0]+s[-5:]
            micro=int(arrPunto[1][:-5])
            dt=datetime.strptime( value, "%Y-%m-%d %H:%M:%S%z" )
            dt=dt+timedelta(microseconds=micro)
            return dtaware_changes_tz(dt, tz_name)
        if format=="JsUtcIso": #2021-08-21T06:27:38.294Z
            if not "Z" in value:
                raise exceptions.CastException(error)
            s=value.replace("T"," ").replace("Z","")

            dtnaive=str2dtnaive(s,"%Y-%m-%d %H:%M:%S.")
            dtaware_utc=dtnaive2dtaware(dtnaive, 'UTC')
            return dtaware_changes_tz(dtaware_utc, tz_name)
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

## epoch is the time from 1,1,1970 in UTC
## return now(timezone(self.name))
def dtaware2epochms(d):
    return d.timestamp()*1000
    
## Return a UTC datetime aware
def epochms2dtaware(n, tz="UTC"):
    utc_unaware=datetime.utcfromtimestamp(n/1000)
    utc_aware=utc_unaware.replace(tzinfo=ZoneInfo('UTC'))#Due to epoch is in UTC
    return dtaware_changes_tz(utc_aware, tz)

## epoch is the time from 1,1,1970 in UTC
## return now(timezone(self.name))
def dtaware2epochmicros(d):
    return int(d.timestamp()*1000000)
## Return a UTC datetime aware
def epochmicros2dtaware(n, tz="UTC"):
    utc_unaware=datetime.utcfromtimestamp(n/1000000)
    utc_aware=utc_unaware.replace(tzinfo=ZoneInfo('UTC'))#Due to epoch is in UTC
    return dtaware_changes_tz(utc_aware, tz)


## Returns a formated string of a dtaware string formatting with a zone name
## @param dt datetime aware object
## @return String
def dtaware2str(value, format="JsUtcIso", ignore_exception=False, ignore_exception_value=None):
    original=value
    allowed=["%Y-%m-%d", "%Y-%m-%d %H:%M:%S", "%Y%m%d %H%M", "%Y%m%d%H%M", "JsUtcIso"]
    error=f"Error in Pydicts.cast.dtaware2str method. Value: {original} Value class: {value.__class__.__name__} Format: {format} Allowed: {allowed}"
    
    if format not in  allowed or value.__class__!=datetime or is_naive(value):
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value
    try:
        if format=="%Y-%m-%d":
            return value.strftime("%Y-%m-%d")
        elif format=="%Y-%m-%d %H:%M:%S": 
            return value.strftime("%Y-%m-%d %H:%M:%S")
        elif format=="%Y%m%d %H%M": 
            return value.strftime("%Y%m%d %H%M")
        elif format=="%Y%m%d%H%M":
            return value.strftime("%Y%m%d%H%M")
        elif format=="JsUtcIso":
            value=dtaware_changes_tz(value, "UTC")
            return value.isoformat().replace("+00:00","Z")
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

## Returns a formated string of a dtaware string formatting with a zone name
## @param dt datetime aware object
## @param format String in ["%Y-%m-%d", "%Y-%m-%d %H:%M:%S", "%Y%m%d %H%M", "%Y%m%d%H%M"]
## @return String
def dtnaive2str(value, format="JsIso", ignore_exception=False, ignore_exception_value=None):
    original=value
    allowed=["%Y-%m-%d", "%Y-%m-%d %H:%M:%S", "%Y%m%d %H%M", "%Y%m%d%H%M", "JsIso"]
    error=f"Error in Pydicts.cast.dtnaive2str method. Value: {original} Value class: {value.__class__.__name__} Format: {format} Allowed: {allowed}"
    if format not in  allowed or value.__class__!=datetime or is_aware(value):
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

    try:
        if format=="%Y-%m-%d":
            return value.strftime("%Y-%m-%d")
        elif format=="%Y-%m-%d %H:%M:%S": 
            return value.strftime("%Y-%m-%d %H:%M:%S")
        elif format=="%Y%m%d %H%M": 
            return value.strftime("%Y%m%d %H%M")
        elif format=="%Y%m%d%H%M":
            return value.strftime("%Y%m%d%H%M")
        elif format=="JsIso":
            return value.strftime("%Y-%m-%dT%H:%M:%S")+"."+str(value.microsecond).zfill(6)
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

## Changes zoneinfo from a dtaware object
## For example:
## - datetime.datetime(2018, 5, 18, 8, 12, tzinfo=<DstTzInfo 'Europe/Madrid' CEST+2:00:00 DST>)
## - libcaloriestrackerfunctions.dtaware_changes_tz(a,"Europe/London")
## - datetime.datetime(2018, 5, 18, 7, 12, tzinfo=<DstTzInfo 'Europe/London' BST+1:00:00 DST>)
## @param dt datetime aware object
## @param tzname String with datetime zone. For example: "Europe/Madrid"
## @return datetime aware object
def dtaware_changes_tz(dt,  tzname):
    if dt==None:
        return None
    if is_aware(dt):
        return dt.astimezone(ZoneInfo(tzname))
    else:
        raise exceptions.CastException(_("A datetime with timezone is needed"))

## Returns a list of tuples (year, month) from a month to another month, both included
## @param year_from Integer
## @param month_from Integer
## @param year_to Integer If none uses current year
## @param month_to Integer If none uses current month
def months(year_from, month_from, year_to=None, month_to=None):
    if year_to is None or month_to is None:
        year_to=date.today().year
        month_to=date.today().month
    r=[]
    end=date_first_of_the_month(year_to, month_to)
    current=date_first_of_the_month(year_from, month_from)
    while True:
        if current>end:
            break
        r.append((current.year,current.month))
        current=date_first_of_the_next_x_months(current.year, current.month, 1)
    return r
    
def timedelta2str(value, ignore_exception=False, ignore_exception_value=None):
    """
        Converts a timedelta object into a string in ISO_8601
    """
    original=value
    error=f"Error in Pydicts.cast.timedelta2str method. Value: {original} Value class: {value.__class__.__name__}"
    if is_noe(value):
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value
            
    try:
        return duration_isoformat(value)
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value
    
def str2timedelta(value, ignore_exception=False, ignore_exception_value=None):
    """
        Converts a string in ISO_8601 in a timedelta or duration
    """
    original=value
    error=f"Error in Pydicts.cast.str2timedelta method. Value: {original} Value class: {value.__class__.__name__}"
    if is_noe(value):
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value
    try:
        return parse_duration(value)
    except:
        if ignore_exception is False:
            raise exceptions.CastException(error)
        else:
            return ignore_exception_value

def is_email(value):
    """
        Returns if a string is a valid email using a regular expression
    """
    if not value.__class__==str:
        return False

    if match(r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$', value):
        return True
    else:
        return False
    



class LodException(Exception):
    pass
    
class CastException(Exception):
    pass
    
class LodXYVException(Exception):
    pass
    
class LodYMVException(Exception):
    pass
    
class LolException(Exception):
    pass

class CurrencyOperationsException(Exception):
    pass



from decimal import Decimal
from pydicts.percentage import Percentage, percentage_between
from pytest import raises

def test_percentage_init():
    assert Percentage(None).value== None
    assert Percentage(12, 1).value == 12
    assert Percentage(12, 0).value== None

def test_percentage_add():
    a=Percentage(1, 2)
    b=Percentage(1, 4)
    assert a+b==Percentage(3, 4)
    assert not a+b==Percentage(None)
    assert Percentage(None)+Percentage(None)==None
    
def test_percentage_sub():
    a=Percentage(1, 2)
    b=Percentage(1, 4)
    assert a-b==Percentage(1, 4)
    assert not a+b==Percentage(None)
    assert Percentage(None)-Percentage(None)==None
    
def test_percentage_neg():
    a=Percentage(1, 2)
    assert -(-a)==a
    assert -Percentage(None)==Percentage(None)
    
def test_percentage_repr():
    assert repr(Percentage(1, 2))=="50.00 %"
    assert repr(Percentage(None))=="None %"
    
def test_percentage_mult():
    assert Percentage(1, 2)*Percentage(1,2 )==Percentage(1, 4)
    assert Percentage(1, 2)*Percentage(None,2 )==Percentage(None)
    with raises(TypeError):
        2*Percentage(1,2 )
    
def test_percentage_truediv():
    assert Percentage(1, 2)/Percentage(1,2 )==Percentage(1, 1)
    assert Percentage(1, 2)/Percentage(None,2 )==Percentage(None)
    
    
def test_percentage_value_100():
    assert Percentage(1, 2).value_100()==Decimal(50)
    assert Percentage(None).value_100()==None
    
def test_percentage_float_100():
    assert Percentage(1, 2).float_100()==50
    assert Percentage(None).float_100()==None
    
def test_percentage_float():
    assert Percentage(1, 2).float()==0.5
    assert Percentage(None).float()==None
    
def test_percentage_lt():
    assert not Percentage(1, 2)<Percentage(1, 4)
    assert Percentage(1, 4)<Percentage(1, 2)
    assert not Percentage(None)<Percentage(1, 2)
    assert not Percentage(1, 2)<Percentage(None)

def test_percentage_isGETZero():
    assert Percentage(1, 2).isGETZero()==True
    assert Percentage(-1, 2).isGETZero()==False
    assert Percentage(None).isGETZero()==False

def test_percentage_isLETZero():
    assert Percentage(1, 2).isLETZero()==False
    assert Percentage(-1, 2).isLETZero()==True
    assert Percentage(None).isLETZero()==False
    
def test_percentage_isGTZero():
    assert Percentage(1, 2).isGTZero()==True
    assert Percentage(-1, 2).isGTZero()==False
    assert Percentage(None).isGTZero()==False

def test_percentage_isLTZero():
    assert Percentage(1, 2).isLTZero()==False
    assert Percentage(-1, 2).isLTZero()==True
    assert Percentage(None).isLTZero()==False

def test_percentage_between():
    assert percentage_between(1, 2)==Percentage(1,1 )
    assert percentage_between(None,1 )==Percentage(None )
    



from pydicts.currency import Currency, currencies_list
from pydicts import lod, exceptions
from pytest import raises
def test_currency():
    assert Currency(None).amount== 0

def test_currency_currencies_dictionary():
    lod.lod_print(currencies_list())
#    assert False
    
    
def test_currency_init():
    assert Currency().string()=="0.00 €"
    
def test_currency_add():
    a=Currency(12.12, "EUR")
    b=Currency(0.88, "USD")
    assert a+a==Currency(24.24, "EUR")
    with raises(exceptions.CurrencyOperationsException):
        a+b
            
def test_currency_eq():
    a=Currency(12.12, "EUR")
    b=Currency(0.88, "USD")
    assert not a==b
    
        
def test_currency_sub():
    a=Currency(12.12, "EUR")
    b=Currency(0.88, "USD")
    assert a-a==Currency(0, "EUR")
    with raises(exceptions.CurrencyOperationsException):
        a-b        
        
def test_currency_mul():
    a=Currency(12.12, "EUR")
    b=Currency(0.88, "USD")
    assert (a*a)==Currency( 146.8944, "EUR")
    with raises(exceptions.CurrencyOperationsException):
        a*b
    with raises(TypeError):
        2*a
    
    a*2==Currency(24.24, "EUR")


def test_currency_lt():
    a=Currency(12.12, "EUR")
    b=Currency(0.88, "USD")
    with raises(exceptions.CurrencyOperationsException):
        a<b
    assert not a<a
    assert a< a+a
    
def test_currency_lte():
    a=Currency(12.12, "EUR")
    b=Currency(0.88, "USD")
    with raises(exceptions.CurrencyOperationsException):
        a<=b
    assert a<=a
    assert a<= a+a

def test_currency_gt():
    a=Currency(12.12, "EUR")
    b=Currency(0.88, "USD")
    with raises(exceptions.CurrencyOperationsException):
        a>b
    assert not a>a
    assert a+a> a
    
def test_currency_gte():
    a=Currency(12.12, "EUR")
    b=Currency(0.88, "USD")
    with raises(exceptions.CurrencyOperationsException):
        a>=b
    assert a>=a
    assert a+a>= a



from gettext import translation
from importlib.resources import files
from pydicts import exceptions
from tabulate import tabulate
        
try:
    t=translation('pydicts', files("pydicts") / 'locale')
    _=t.gettext
except:
    _=str


## LOR is a list of list. 
## @param rows LOR
## @param index int with the index of the position where we are going to insert row
## @param column List with the values to add. Must be of the same size of rows
def lol_add_column(rows, index, column):
    if len(rows)!=len(column):
        raise exceptions.LolException(_("I can't add a column with different size of lol"))
    r_rows=[]
    for i, row in enumerate(rows):
        r_rows.append(row[0:index] + [column[i],] + row[index:len(row)])
    return r_rows

def lol_add_row(lol_, index, row):
    """
        Adds a row to a list of rows
        Checks if new row has the same size of the first row of the LOL
        Parameters:
            - lol_
            - index: int with the index of the position where we are going to insert row
            - row
        Returns
            LOL
            
        Will raise error:
            - lol.lol_add_row(lol_, 1, [2,3]) #New row size is different to first row
    """
    if len(lol_)>0 and len(lol_[0])!=len(row):
        raise exceptions.LolException(_("I can't add a row with different size of the first row of the lol"))
    lol_.insert(index, row)
    return lol_

## Returns a list with object in positions removed
def list_remove_positions(l, listindex):
    if l is None:
        raise exceptions.LolException(_("I can't remove positions from a None list"))
    r=[]
    for i, o in enumerate(l):
        if i not in listindex:
            r.append(o)
    return r

## Order data columns. 
def lol_order_by(lol_, index, reverse=False, none_at_top=True):
    """
        Orders a list of lists by a index
        Parameters:
            - lol
            - index
            - reverse
            - none_at_top: None values are set at the beginning
    """
    
    nonull=[]
    null=[]
    for o in lol_:
        com=o[index]
        if com is None:
            null.append(o)
        else:
            nonull.append(o)
    nonull=sorted(nonull, key=lambda item: item[index], reverse=reverse)
    if none_at_top==True:#Set None at top of the list
        return null+nonull
    else:
        return nonull+null


## lol is a list of list. Naned List Of Rows, used in myqtablewidget
## @param list_of_indexes is a list of column indexes to remove
def lol_remove_columns(rows, list_of_indexes):
    r_rows=[]
    for i, row in enumerate(rows):
        r_rows.append(list_remove_positions(row,list_of_indexes))
    return r_rows

## lol is a list of list. Naned List Of Rows, used in myqtablewidget
## @param listindex is a list of column indexes to remove
def lol_remove_rows(rows, listindex):
    return list_remove_positions(rows, listindex) #It's a list but of row

## Return a lol transposed. Changed rows by columns
def lol_transposed(lol):
    if lol is None:
        raise exceptions.LolException(_("I can't traspaso a None object"))
    if len(lol)==0:
        return []
    r=[]
    columns=len(lol[0])
    for column in range(columns):
        tran_row=[]
        for row in lol:
            tran_row.append(row[column])
        r.append(tran_row)
    return r

## Extract a column from the list of row
def lol_get_row(lol, row):
    return lol[row]

## Extract a column from the list of row
def lol_get_column(lol, column):
    r=[]
    for row in lol:
        r.append(row[column])
    return r

## Return sum of values of a column from and index to and other index position. This method ignores None values
## This method can sum several objects
## @param row
## @param from_index
## @param to_index
## @param zerovalue 0 or Money(self.mem, 0, self.mem.localcurrency)....
def lol_sum_row(row, from_index=0, to_index=None, zerovalue=0):
    s=zerovalue
    for i, column in enumerate(row):
        if i>=from_index and i<=to_index:
            if column is not None:
                s=s + column
    return s

## Return sum of values of a column from and index to and other index position. This method ignores None values
## This method can sum several objects
## @zerovalue 0 or Money(self.mem, 0, self.mem.localcurrency)....
def lol_sum_column(lol, column, from_index, to_index, zerovalue=0):
    s=zerovalue
    for i, row in enumerate(lol):
        if i>=from_index and i<=to_index:
            if row[column] is not None:
                s=s + row[column]
    return s

def lol_print(lol_, number=None):
    """
        Prints a list of list with tabulate module.
        
        Parameters:
            - lol_: List of list
            - number: Númber of lists to print. If None prints all lod. (defaults to None)
    """
    number=len(lol_) if number is None else number

    if len(lol_)==0:
        print(_("lol_print: This list of lists hasn't data to print"))
        return
    if number==0:
        print(_("lol_print: No data was printed due to you selected 0 rows"))
        return
    print(tabulate(lol_[0:number], tablefmt="psql"))



from datetime import date, time, datetime, timedelta
from decimal import Decimal
from pydicts import casts, exceptions
from pytest import raises
from zoneinfo import ZoneInfo

def test_is_noe():
    assert casts.is_noe(None)==True
    assert casts.is_noe(1)==False
    assert casts.is_noe("")==True
    assert casts.is_noe("HELLO")==False    
    
def test_none2alternative():
    assert casts.none2alternative(1, 1)==1
    assert casts.none2alternative(None, Decimal(0))==Decimal(0)
    assert casts.none2alternative(None, None)==None
    
def test_object_or_empty():
    assert casts.object_or_empty(None)==""
    assert casts.object_or_empty(1)==1
    assert casts.object_or_empty("")==""

def test_str2decimal():
    with raises(exceptions.CastException):
        assert casts.str2decimal(None)
    assert casts.str2decimal("", ignore_exception=True)==None
    
    with raises(exceptions.CastException):
        assert casts.str2decimal("")
    assert casts.str2decimal("", ignore_exception=True)==None
    
    with raises(exceptions.CastException):
        assert casts.str2decimal("2.123,25")
    assert casts.str2decimal("2.123,25", ignore_exception=True)==None
    
    with raises(exceptions.CastException):
        assert casts.str2decimal("2.123,25")
    assert casts.str2decimal("2.123,25", ignore_exception=True)==None
    
    with raises(exceptions.CastException):
        assert casts.str2decimal("2,123.25")
    assert casts.str2decimal("2,123.25", ignore_exception=True)==None
    
    assert casts.str2decimal("121212.123")==Decimal("121212.123")

def test_str2bool():
    with raises(exceptions.CastException):
        assert casts.str2bool(None)
    casts.str2bool(None, ignore_exception=True)==None
    
    with raises(exceptions.CastException):
        assert casts.str2bool(1)
    assert casts.str2bool(1, ignore_exception=True)==None
    
    with raises(exceptions.CastException):
        assert casts.str2bool(0)
    assert casts.str2bool(0, ignore_exception=True)==None
    assert casts.str2bool("1")==True
    assert casts.str2bool("0")==False
    with raises(exceptions.CastException):
        assert casts.str2bool(True)
    with raises(exceptions.CastException):
        assert casts.str2bool(False)
    assert casts.str2bool("true")==True
    assert casts.str2bool("True")==True
    assert casts.str2bool("false")==False
    assert casts.str2bool("False")==False
    

def test_date_first_of_the_next_x_months():
    date_plus_3_months=casts.date_first_of_the_next_x_months(2023, 11, 3)
    assert date_plus_3_months.year==2024
    assert date_plus_3_months.month==2
    assert date_plus_3_months.day==1
    date_minus_12_months=casts.date_first_of_the_next_x_months(2023, 11, -12)
    assert date_minus_12_months.year==2022
    assert date_minus_12_months.month==11
    assert date_minus_12_months.day==1
    
def test_date_last_of_the_next_x_months():
    date_plus_3_months=casts.date_last_of_the_next_x_months(2023, 11, 3)
    assert date_plus_3_months.year==2024
    assert date_plus_3_months.month==2
    assert date_plus_3_months.day==29
    date_minus_12_months=casts.date_last_of_the_next_x_months(2023, 11, -12)
    assert date_minus_12_months.year==2022
    assert date_minus_12_months.month==11
    assert date_minus_12_months.day==30

def test_date_first_of_the_year():
    assert casts.date_first_of_the_year(2022)==date(2022, 1, 1)
    
def test_date_last_of_the_year():
    assert casts.date_last_of_the_year(2023)==date(2023, 12, 31)

def test_bytes2str():
    with raises(exceptions.CastException):
        assert casts.bytes2str(None)==None
    assert casts.bytes2str(None, ignore_exception=True)==None
    assert casts.bytes2str(b"Hello")=="Hello"
    
def test_str2bytes():
    with raises(exceptions.CastException):
        assert casts.str2bytes(None)==None
    assert casts.str2bytes(None, ignore_exception=True)==None
    assert casts.str2bytes("Hello")==b"Hello"

def test_is_aware():
    assert casts.is_aware(casts.dtnaive_now())==False
    assert casts.is_aware(casts.dtaware_now())==True

def test_is_naive():
    assert casts.is_naive(casts.dtnaive_now())==True
    assert casts.is_naive(casts.dtaware_now())==False

def test_dtaware():
    dtaware_utc=casts.dtaware_now()
    assert dtaware_utc==casts.dtaware(dtaware_utc.date(), dtaware_utc.time(), "UTC")
    
def test_dtaware2dtnaive():
    dt_naive=  datetime(2023, 11, 26, 17, 5, 5, 123456)
    dt_aware=casts.dtnaive2dtaware(dt_naive, "UTC")
    with raises(exceptions.CastException):
        casts.dtaware2dtnaive(dt_naive)
    assert dt_naive==casts.dtaware2dtnaive(dt_aware)

def test_dtnaive():
    assert casts.dtnaive(date(2020,1 , 1), time(1, 1, 1, 1))==datetime(2020, 1, 1, 1, 1, 1, 1)
    with raises(ValueError):
        casts.dtnaive(date(2020,1 , 61), time(1, 1, 1, 1))

def test_dtnaive2dtaware():
    naive=datetime(2023, 1, 1, 0, 0, 0, 0)
    assert casts.dtnaive2dtaware(naive, "Europe/Madrid")==naive.replace(tzinfo=ZoneInfo("Europe/Madrid"))

def test_date_first_of_the_month():
    assert casts.date_first_of_the_month(2023, 11)==date(2023, 11, 1)
    with raises(ValueError):
        casts.date_first_of_the_month(2023, 13)

def test_date_last_of_the_month():
    assert casts.date_last_of_the_month(2023, 11)==date(2023, 11, 30)
    with raises(ValueError):
        casts.date_first_of_the_month(2023, 13)

def test_dtaware_now():
    pass
    
def test_dtnaive_now():
    pass
    
def test_dtaware_month_start():
    pass

def test_dtaware_month_end():
    assert casts.dtaware_month_end(2023, 11, "UTC")==datetime(2023, 11, 30, 23, 59, 59, 999999, ZoneInfo("UTC"))
    
def test_dtaware_year_start():
    assert casts.dtaware_year_start(2023, "UTC")==datetime(2023, 1, 1,0, 0, 0, 0, ZoneInfo("UTC"))

def test_dtaware_year_end():
    assert casts.dtaware_year_end(2023, "UTC")==datetime(2023, 12, 31, 23, 59, 59, 999999, ZoneInfo("UTC"))

def test_dtnaive_day_end():
    assert casts.dtnaive_day_end(casts.dtnaive_now()).time()==time(23, 59, 59, 999999)

def test_dtnaive_day_end_from_date():
    assert casts.dtnaive_day_end_from_date(date(2023, 12, 29))==datetime(2023, 12, 29, 23, 59, 59, 999999)

def test_dtaware_day_end():
    pass

def test_dtaware_day_end_from_date():
    assert casts.dtaware_day_end_from_date(date(2023, 12, 29), "UTC")==datetime(2023, 12, 29, 23, 59, 59, 999999, ZoneInfo("UTC"))

def test_dtaware_day_start():
    pass

def test_dtaware_day_start_from_date():
    pass

def test_str2time():
#    allowed=["HH:MM", "HH:MM:SS","HH:MMxx"]
    with raises(exceptions.CastException):
        assert casts.str2time(None)
    assert casts.str2time("09:05", "HH:MM")==time(9, 5)
    assert casts.str2time("09:05:54", "HH:MM:SS")==time(9, 5, 54)
    assert casts.str2time("09:05:54.123")==time(9, 5, 54, 123000)
    assert casts.str2time("09:05:54.000123")==time(9, 5, 54, 123)
    assert casts.str2time("09:05pm", "HH:MMxx")==time(21, 5)
    
    
def test_str2timedelta():
    assert casts.str2timedelta("P396DT01H01M01.000001S")==timedelta(days=396,  hours=1,  minutes=1, seconds=1, microseconds=1)
    assert casts.str2timedelta("PT0S")==timedelta( seconds=0)
    
    
    with raises(exceptions.CastException):
        casts.str2timedelta(None)
    with raises(exceptions.CastException):
        casts.str2timedelta("")
    
    assert casts.str2timedelta(None, ignore_exception=True)==None
    assert casts.str2timedelta("", ignore_exception=True)==None
    
def test_timedelta2str():
    assert casts.timedelta2str(timedelta(days=396,  hours=1,  minutes=1, seconds=1, microseconds=1))=="P396DT1H1M1.000001S"
    assert casts.timedelta2str(timedelta(days=0))=="P0D"
    
    with raises(exceptions.CastException):
        casts.timedelta2str(None)
    with raises(exceptions.CastException):
        casts.timedelta2str("")
        
        
    assert casts.timedelta2str(None, ignore_exception=True)==None
    assert casts.timedelta2str("", ignore_exception=True)==None

def test_time2str():
    with raises(exceptions.CastException):
        assert casts.time2str(None)
        
    assert casts.time2str(None, ignore_exception=True)==None
    time_=time(9, 5, 54, 123456)
    assert casts.time2str(time_, "HH:MM")=="09:05"
    assert casts.time2str(time_, "HH:MM:SS")=="09:05:54"
    assert casts.time2str(time_)=="09:05:54.123456"

def test_str2date():
    with raises(exceptions.CastException):
        assert casts.str2date(None)
    assert casts.str2date(None, ignore_exception=True)==None
    
    with raises(exceptions.CastException):
        assert casts.str2date("2023")
        
    with raises(exceptions.CastException):
        assert casts.str2date(2023)
    
    assert casts.str2date("2023-11-26")==date(2023, 11, 26)
    assert casts.str2date("2023-11-26", "YYYY-MM-DD")==date(2023, 11, 26)
    assert casts.str2date("26/11/2023", "DD/MM/YYYY")==date(2023, 11, 26)
    assert casts.str2date("26.11.2023", "DD.MM.YYYY")==date(2023, 11, 26)
    assert casts.str2date("26/11", "DD/MM")==date(date.today().year, 11, 26)

def test_str2dtnaive():
    assert casts.str2dtnaive("202311261705", "%Y%m%d%H%M")==datetime(2023, 11, 26, 17, 5)
    assert casts.str2dtnaive("2023-11-26T17:05:05")==datetime(2023, 11, 26, 17, 5, 5)
    assert casts.str2dtnaive("2023-11-26T17:05:05.123456", "JsIso")==datetime(2023, 11, 26, 17, 5, 5, 123456)

    with raises(exceptions.CastException):
        casts.str2dtnaive("2023-11-26T17:05:05.123456Z")
    
    assert casts.str2dtnaive("2023-11-26T17:05:05.123456Z", ignore_exception=True)==None


def test_str2dtaware():   
    with raises(exceptions.CastException):
        casts.str2dtaware("2023-11-26T17:05:05Z", "YYYY-mm-dd")
    
    assert casts.str2dtaware("2023-11-26T17:05:05Z", "YYYY-mm-dd", ignore_exception=True)==None
    assert casts.str2dtaware("2023-11-26T17:05:05Z")==datetime(2023, 11, 26, 17, 5, 5, tzinfo=ZoneInfo('UTC'))
    assert casts.str2dtaware("2023-11-26T17:05:05", "JsUtcIso", ignore_exception=True)==None

def test_dtaware2str():    
    dt_naive=  datetime(2023, 11, 26, 17, 5, 5, 123456)
    dt_aware=casts.dtnaive2dtaware(dt_naive, "UTC")
    with raises(exceptions.CastException):
        casts.dtaware2str(dt_naive)
    assert casts.dtaware2str(dt_aware, "%Y-%m-%d")=="2023-11-26"
    assert casts.dtaware2str(dt_aware, "%Y-%m-%d %H:%M:%S")=="2023-11-26 17:05:05"
    assert casts.dtaware2str(dt_aware, "%Y%m%d %H%M")=="20231126 1705"
    assert casts.dtaware2str(dt_aware, "%Y%m%d%H%M")=="202311261705"
    assert casts.dtaware2str(dt_aware, "JsUtcIso")=="2023-11-26T17:05:05.123456Z"


def test_dtnaive2str():
    dt_naive=  datetime(2023, 11, 26, 17, 5, 5, 123456)
    dt_aware=casts.dtnaive2dtaware(dt_naive, "UTC")
    with raises(exceptions.CastException):
        casts.dtnaive2str(dt_aware)
    
    assert casts.dtnaive2str(dt_naive, "%Y-%m-%d")=="2023-11-26"
    assert casts.dtnaive2str(dt_naive, "%Y-%m-%d %H:%M:%S")=="2023-11-26 17:05:05"
    assert casts.dtnaive2str(dt_naive, "%Y%m%d %H%M")=="20231126 1705"
    assert casts.dtnaive2str(dt_naive, "%Y%m%d%H%M")=="202311261705"
    assert casts.dtnaive2str(dt_naive, "JsIso")=="2023-11-26T17:05:05.123456"

def test_dtaware_changes_tz():
    #Sacado date en linux
    dt_madrid=datetime(2023, 11, 26, 12, 0,  tzinfo=ZoneInfo("Europe/Madrid"))
    dt_utc=datetime(2023, 11, 26, 11, 0,  tzinfo=ZoneInfo("UTC"))
    
    assert casts.dtaware_changes_tz(dt_madrid, "UTC")==dt_utc
    assert casts.dtaware_changes_tz(dt_utc, "Europe/Madrid")==dt_madrid

def test_months():
    assert casts.months(2023, 11, 2024, 1)== [(2023, 11), (2023, 12), (2024, 1)]
    assert casts.months(2023, 9,  2023, 9)== [(2023, 9)]

def test_bytes2base64bytes():
    s="Elvis Presley"
    bytes=casts.str2bytes(s)
    base64bytes=casts.bytes2base64bytes(bytes)
    bytes2=casts.base64bytes2bytes(base64bytes)
    assert s==casts.bytes2str(bytes2)
    
def test_base64bytes2bytes():
    pass

def test_dtaware2epochmicros():
    pass
    
def test_epochmicros2dtaware():
    pass

def test_epochms2dtaware():
    pass

def test_email():
    assert casts.is_email(None)==False
    assert casts.is_email("")==False
    assert casts.is_email(12)==False
    assert casts.is_email("hi.hi.com")==False
    assert casts.is_email("hi@hi@.com")==False
    assert casts.is_email("hi@hi.com.")==False
    assert casts.is_email("hi@hi.comm")
    
    
    
    



from pydicts import __version__, lod
from os import system

def release():
    print("""Nueva versión:
  * Cambiar la version en pyproject.toml
  * Cambiar la versión y la fecha en __init__.py
  * Ejecutar otra vez poe release
  * git checkout -b pydicts-{0}
  * Modificar el Changelog en jupyter/changelog.md
  * poe translate
  * linguist
  * poe translate
  * poe test
  * poe jupyter
  * git commit -a -m 'pydicts-{0}'
  * git push
  * Hacer un pull request con los cambios a main
  * Hacer un nuevo tag en GitHub
  * git checkout main
  * git pull
  * poetry build
  * poetry publish
  * Crea un nuevo ebuild de pydicts en Gentoo con la nueva versión
  * Subelo al repositorio myportage

""".format(__version__))

def jupyter():
    """
        This command is for developing
        This page is published into https://turulomio.github.io/pydicts/ with github actions
    """
    system("jupyter-book build jupyter")
    system("rm -Rf docs")
    system("mv jupyter/_build/html docs/")
    system("touch docs/.nojekyll")
    
def coverage():
    system("coverage run --omit='*uno.py' -m pytest && coverage report && coverage html")


def translate():
    #es
    system("xgettext -L Python --no-wrap --no-location --from-code='UTF-8' -o pydicts/locale/pydicts.pot pydicts/*.py")
    system("msgmerge -N --no-wrap -U pydicts/locale/es.po pydicts/locale/pydicts.pot")
    system("msgfmt -cv -o pydicts/locale/es/LC_MESSAGES/pydicts.mo pydicts/locale/es.po")
    system("msgfmt -cv -o pydicts/locale/en/LC_MESSAGES/pydicts.mo pydicts/locale/en.po")
    
    
def currencies():
    """
        Due to slow ccy updates to python new releases I decided to covert currency list to a updatable json file with this scripts
        
        Currency list is obtained from https://github.com/quantmind/ccy/ project this is its licence:
        
        YOU NEED TO INSTALL CCY MODULE MANUALLY  WHEN YOU WANT TO UPDATE currencies.json FILE (pip install ccy)
        
        Copyright (c) 2009-2023, Quantmind
        All rights reserved.

        Redistribution and use in source and binary forms, with or without modification,
        are permitted provided that the following conditions are met:

         * Redistributions of source code must retain the above copyright notice,
           this list of conditions and the following disclaimer.
         * Redistributions in binary form must reproduce the above copyright notice,
           this list of conditions and the following disclaimer in the documentation
           and/or other materials provided with the distribution.
         * Neither the name of the author nor the names of its contributors
           may be used to endorse or promote products derived from this software without
           specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
        ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
        WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
        IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
        INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
        BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
        DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
        LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
        OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
        OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    from ccy import dump_currency_table
    from html import unescape
    lod_=[]
    for code_, name, iso, symbol, country, order, rounding in dump_currency_table()[1:]:
        print(code_, name, iso, symbol, country, order, rounding)
        lod_.append({
            "code":code_, 
            "name":name, 
            "iso":iso, 
            "symbol_html":symbol, 
            "symbol": unescape(symbol), 
            "country":country, 
            "order":order, 
            "rounding":rounding
        })
    lod_.append({
        "code":"u", 
        "name": "Unit", 
        "iso":"", 
        "symbol_html":"u", 
        "symbol": "u", 
        "country":"WW", 
        "order": 0, 
        "rounding":6, 
    })
    lod_=lod.lod_order_by(lod_,"code")
    with open("pydicts/currencies.json", "w") as f:
        f.write(str(lod_))
        



from datetime import datetime, date
from decimal import Decimal
from pydicts import lod
from pytest import raises, fixture

empty_lod=[]

@fixture(autouse=True)
def reload_lod_():
    global lod_
    lod_=[]
    lod_.append({"a": datetime.now(), "b": date.today(), "c": Decimal('12.32'), "d": None, "e": int(12), "f":None, "g":True, "h":False})
    lod_.append({"a": datetime.now(), "b": date.today(), "c": Decimal('-12.32'), "d": 16, "e": int(12), "f":None, "g":True, "h":False})


def tests_dictkv():
    d=lod.lod2dictkv(lod_, "a","b")
    assert list(d.values())[0]==date.today()
    

def tests_lod_has_key():
    assert lod.lod_has_key(lod_,"c")==True
    assert lod.lod_has_key(lod_,"cc")==False
    assert lod.lod_has_key(empty_lod, "c")==False
    
def test_remove_duplicates():
    lod_=[
        {"a":1,  "b":2}, 
        {"a":1,  "b":2}, 
        {"a":1,  "b":2}, 
    ]
    lod_=lod.lod_remove_duplicates(lod_)
    assert len(lod_)==1
    
def tests_lod_sum():
    assert lod.lod_sum(lod_, "c")==0
    assert lod.lod_sum(lod_, "d")==16
    with raises(TypeError) as excinfo:
        lod.lod_sum(lod_, "d", ignore_nones=False)==16
    assert "NoneType" in str(excinfo.value)
    
def test_lod_average():
    lod.lod_print(lod_)
    assert lod.lod_average(lod_,"c")==0
    assert lod.lod_average(lod_,"d")==8

def test_lod_average_ponderated():
    assert lod.lod_average_ponderated(lod_,"e", "c")==0
    
def test_lod_median():
    assert lod.lod_median(lod_,"c")==0
    
def tests_lod_sum_positives():
    assert lod.lod_sum_positives(lod_, "c")==Decimal("12.32")
    assert lod.lod_sum_positives(lod_, "d")==16
    
def tests_lod_sum_negatives():
    assert lod.lod_sum_negatives(lod_, "c")==Decimal("-12.32")
    assert lod.lod_sum_negatives(lod_, "d")==0
    
def test_lod_remove_key():
    assert lod.lod_has_key(lod_, "d")
    lod.lod_remove_key(lod_, "d")
    assert not lod.lod_has_key(lod_, "d")
    
def test_lod_filter_keys():
    new_lod=lod.lod_filter_keys(lod_,  ["g", "h"])
    assert "a" not in lod.lod_keys(new_lod)
    assert "g" in lod.lod_keys(new_lod)
    assert "h" in lod.lod_keys(new_lod)


def test_lod_filter_dictionaries():
    #Filtering by index
    new_lod=lod.lod_filter_dictionaries(lod_, lambda d,  index: index==1)
    lod.lod_print(new_lod)
    assert len(new_lod)==1
    
    #Filtering by d
    new_lod=lod.lod_filter_dictionaries(lod_, lambda d,  index: d["g"] is None)
    assert len(new_lod)==0

def test_calculate_clone():
    new_lod=lod.lod_calculate(lod_, "NEW C",   lambda d,  index: d['c']*100, clone=True)
    lod.lod_print(new_lod)
    assert "NEW C" in lod.lod_keys(new_lod)
    assert "NEW C" not in lod.lod_keys(lod_)
    
def test_calculate():
    new_lod=lod.lod_calculate(lod_, "NEW C",   lambda d,  index: d['c']*100)
    lod.lod_print(new_lod)
    assert "NEW C" in lod.lod_keys(new_lod)
    assert "NEW C" in lod.lod_keys(lod_)
    
def test_lod2dod():
    dod_=lod.lod2dod(lod_, "c")
    print(dod_)
    assert dod_[Decimal("12.32")]==lod_[0]

def test_dod2lod():
    dod_=lod.lod2dod(lod_, "c")
    assert lod_==lod.dod2lod(dod_)

def test_count():
    assert lod.lod_count(lod_,lambda d, index: d["c"]>0)==1, "Error counting"
    assert lod.lod_count(lod_,lambda d, index: d["f"] is None)==2, "Error counting"
    assert lod.lod_count(lod_,lambda d, index: index>0)==1, "Error counting"


from pydicts import lol, exceptions
from pytest import raises,  fixture

@fixture(autouse=True)
def reload_lol_():
    global lol_
    lol_=[]
    for i in range(10):
        lol_.append([1*i,2*i,3*i])

def test_lol_add_column():
    new_lol=lol.lol_add_column(lol_, 2,  [1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
    assert len(lol_[0])+1==len(new_lol[0])
    with raises(exceptions.LolException):
        lol.lol_add_column(lol_, 2,  [1, ]) #Different size

def test_lol_order_by():
    new_lol=lol.lol_add_row(lol_, 2,  [None]*3)
    new_lol=lol.lol_order_by(new_lol, 2)
    assert new_lol[0][0]==None
    new_lol=lol.lol_order_by(new_lol, 2, reverse=True, none_at_top=False)
    assert new_lol[0][0]==9

def test_lol_remove_columns():
    new_lol=lol.lol_remove_columns(lol_, [2])
    assert len(lol_[0])-1==len(new_lol[0])

def test_lol_print():
    lol.lol_print(lol_)
    lol.lol_print(lol_,0)
    lol.lol_print(lol_,-1)
    
def test_lol_transposed():
    transposed=lol.lol_transposed(lol_)
    assert transposed[0]==[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    transposed=lol.lol_transposed([])
    assert transposed==[]

    with raises(exceptions.LolException):
        lol.lol_transposed(None)

#
#    lor=[]
#    column_to_add=[]
#    for i in range(10):
#        lor.append([1*i,2*i,3*i])
#        column_to_add.append(-i)
#    print_lor(lor)
#
#    lor=lor_add_column(lor, 0, column_to_add)
#    lor=lor_add_column(lor, 2, column_to_add)
#    lor=lor_add_column(lor, 5, column_to_add)
#    print_lor(lor)
#
#    a=lor_remove_columns(lor,[2,3])
#    print_lor(a)
#    b=lor_remove_rows(a,[8,9])
#    print_lor(b)
#
#    c=lor_transposed(b)
#    print_lor(c)
#
#    d=Decimal("12.3")
#    json_d=var2json(d)
#    print (d, json_d, json_d.__class__)
#    d=None
#    json_d=var2json(d)
#    print (d, json_d, json_d.__class__)
#    d=True
#    json_d=var2json(d)
#    print (d, json_d, json_d.__class__)



from gettext import translation
from importlib.resources import files
from pprint import PrettyPrinter
        
try:
    t=translation('pydicts', files("pydicts") / 'locale')
    _=t.gettext
except:
    _=str

def dod_print(dod_, indent=4, depth=None, width=80, sort_dicts=True):
    """
        Make a nice print of a dictionary of nested dictionaries
    """
    pp=PrettyPrinter(indent=4, compact=False, depth=None,  width=width, sort_dicts=sort_dicts)
    pp.pprint(dod_)
    



# from shutil import which
# from os import remove

# if which("pdflatex") is not None:
#     from pydicts import pylatex
#     from pylatex import Document, Section, Subsection
#     from pylatex.package import Package
#     from pylatex.utils import italic, NoEscape
#     lod=[
#             {"year": 2022, "month": 1, "my_sum": 12},
#             {"year": 2021, "month": 2, "my_sum": 123},
#             {"year": 2019, "month": 5, "my_sum": 1},
#             {"year": 2022, "month": 12, "my_sum": 12},
#     ]
#     def tests_pylatex_table_header():

#         doc = Document()
#         doc.packages.append(Package('xcolor'))

#         with doc.create(Section('A section')):
#             doc.append('Some regular text and some ')
#             doc.append(italic('italic text. '))

#             with doc.create(Subsection('A subsection')):
#                 doc.append('Also some crazy characters: $&#{}')
                
#         doc.append(NoEscape("\\centering"))
#         pylatex.pylatex_table(doc, lod)
#         pylatex.pylatex_table(doc, [])
#         pylatex.pylatex_table_with_matched_values(doc,  [2022, 2, 12], lod, code_="|l|c|r|", match_color="teal", unmatch_color="red")
#         doc.generate_pdf('test_pylatex_table_header', clean_tex=False)
#         remove("test_pylatex_table_header.pdf")
#         remove("test_pylatex_table_header.tex")



from colorama import Style,Fore
from pydicts.currency import Currency
from pydicts.percentage import Percentage

def blue(s):
    """
        Shows a string in blue bright color
    """
    return Style.BRIGHT + Fore.BLUE + str(s) + Style.RESET_ALL

def red(s):
    """
        Shows a string in red bright color
    """
    return Style.BRIGHT + Fore.RED + str(s) + Style.RESET_ALL

def green(s):
    """
        Shows a string in green bright color
    """
    return Style.BRIGHT + Fore.GREEN + str(s) + Style.RESET_ALL

def magenta(s):
    """
        Shows a string in magenta bright color
    """
    return Style.BRIGHT + Fore.MAGENTA + str(s) + Style.RESET_ALL

def yellow(s):
    """
        Shows a string in yellow bright color
    """
    return Style.BRIGHT + Fore.YELLOW + str(s) + Style.RESET_ALL

def white(s):
    """
        Shows a string in white bright color
    """
    return Style.BRIGHT + str(s) + Style.RESET_ALL

def cyan(s):
    """
        Shows a string in cyan bright color
    """
    return Style.BRIGHT + Fore.CYAN + str(s) + Style.RESET_ALL

def currency_color(value, currency_):
    """
        Shows a currency string in green or red color
    """
    if value>=0:
        return green(Currency(value,currency_))
    else:
        return red(Currency(value,currency_))

def percentage_color(value):
    """
        Shows a percentage string in green or red color
    """
    if value>=0:
        return green(Percentage(value,1))
    else:
        return red(Currency(value,1))

def value_color(value):
    """
        Shows a value string in green or red color
    """
    if value>=0:
        return green(value)
    else:
        return red(value)



from datetime import date
from pydicts import lod
from gettext import translation
from importlib.resources import files

try:
    t=translation('pydicts', files("pydicts") / 'locale')
    _=t.gettext
except:
    _=str
"""
    lod_ymv example:
    +------------------+-------------------+---------+
|   datetime__year |   datetime__month |   quote |
|------------------+-------------------+---------|
|             2020 |                 1 |   82.92 |
|             2020 |                 2 |   67.41 |
|             2020 |                 3 |   51.24 |
|             2020 |                 4 |   61.98 |
|             2020 |                 5 |   65.21 |
|             2020 |                 6 |   66.69 |
|             2020 |                 7 |   70.55 |
|             2020 |                 8 |   81.31 |
|             2020 |                 9 |   76.6  |
|             2020 |                10 |   74.33 |
|             2020 |                11 |   85.59 |
|             2020 |                12 |   89.84 |
|             2021 |                 1 |   90.52 |
|             2021 |                 2 |   96.19 |
|             2021 |                 3 |  106.42 |
|             2021 |                 4 |  114.54 |
|             2021 |                 5 |  113.42 |
|             2021 |                 6 |  122.66 |
|             2021 |                 7 |  128.38 |
|             2021 |                 8 |  137.18 |
|             2021 |                 9 |  128.82 |
|             2021 |                10 |  143.94 |
|             2021 |                11 |  147.06 |
|             2021 |                12 |  159.98 |
|             2022 |                 1 |  139.62 |
|             2022 |                 2 |  134    |
|             2022 |                 3 |  148.32 |
|             2022 |                 4 |  131.9  |
|             2022 |                 5 |  122.38 |
|             2022 |                 6 |  104.96 |
|             2022 |                 7 |  125.88 |
|             2022 |                 8 |  120.36 |
|             2022 |                 9 |  103.98 |
|             2022 |                10 |  114.4  |
|             2022 |                11 |  113.64 |
|             2022 |                12 |  102.12 |
|             2023 |                 1 |  111.62 |
|             2023 |                 2 |  110.42 |
|             2023 |                 3 |  112.76 |
|             2023 |                 4 |  113.96 |
+------------------+-------------------+---------+

"""


## Converts a tipical groyp by lor with year, month, value into an other lor with year, 1, 2, 3 .... 12, total 
def lod_ymv_transposition(ld, key_year="year", key_month="month", key_value="value"):
    """
        Covert lod_ymv into a trasposition form 
        +--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+---------+
        |   year |     m1 |     m2 |     m3 |     m4 |     m5 |     m6 |     m7 |     m8 |     m9 |    m10 |    m11 |    m12 |   total |
        |--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+---------|
        |   2020 |  82.92 |  67.41 |  51.24 |  61.98 |  65.21 |  66.69 |  70.55 |  81.31 |  76.6  |  74.33 |  85.59 |  89.84 |  873.67 |
        |   2021 |  90.52 |  96.19 | 106.42 | 114.54 | 113.42 | 122.66 | 128.38 | 137.18 | 128.82 | 143.94 | 147.06 | 159.98 | 1489.11 |
        |   2022 | 139.62 | 134    | 148.32 | 131.9  | 122.38 | 104.96 | 125.88 | 120.36 | 103.98 | 114.4  | 113.64 | 102.12 | 1461.56 |
        |   2023 | 111.62 | 110.42 | 112.76 | 113.96 |   0    |   0    |   0    |   0    |   0    |   0    |   0    |   0    |  448.76 |
        +--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+---------+
    """
    if len(ld)==0:
       return []

    if not key_year in ld[0] or not key_month in ld[0] or not key_value in ld[0]:
        print(_("Keys names are not correct in dictionary in lod_ymv_transposition function"))
        return None

    min_year=lod.lod_min_value(ld, key_year)
    max_year=lod.lod_max_value(ld, key_year)
    #Initialize result
    r=[]
    for year in range(min_year,max_year+1):
        r.append({"year": year, "m1":0, "m2":0,  "m3":0, "m4":0, "m5":0, "m6":0, "m7":0, "m8":0, "m9":0, "m10":0, "m11":0, "m12":0, "total":0})

    #Assign values
    for d in ld:
        r[d[key_year]-min_year]["m"+str(d[key_month])]=r[d[key_year]-min_year]["m"+str(d[key_month])]+d[key_value]

    #Calculate totals
    for year in range(min_year,max_year+1):
        d=r[year-min_year]
        d["total"]=d["m1"]+d["m2"]+d["m3"]+d["m4"]+d["m5"]+d["m6"]+d["m7"]+d["m8"]+d["m9"]+d["m10"]+d["m11"]+d["m12"]

    return r
    
def is_noz(v):
    if v is None:
        return True
    if v==0:
        return True
    return False
    
def ymv_transposition_first_value_not_noz(lod_ymv_transposition):
    """
        REturns a dict with the coords of the first value of the ymv_transposition not null
        noz null or zero
        
    """
    for d in lod_ymv_transposition:
        for i in range(1, 13):
            if not is_noz(d[f"m{i}"]):
                return {"year": d["year"],  "month": i ,  "value":d[f"m{i}"]  }
    return None
    
def d_ymv_transposition_first_key_not_noz(d):
    """
        REturns the first key mX not null in dictionary
        noz null or zero
    """
    r=None
    for i in range(1, 13):
        if not is_noz(d[f"m{i}"]):
            r= f"m{i}"
            break
    return r

def d_ymv_transposition_last_key_not_noz(d):
    """
        REturns the last key mX not null in dictionary
        noz null or zero
    """
    r=None
    for i in reversed(range(1, 13)):
        if not is_noz(d[f"m{i}"]):
            r= f"m{i}"
            break
    return r

    
def lod_ymv_transposition_with_percentages(lod_ymv_transposition):
    """
        Replace a lod_ymv_transposition values to Percentages
    """
    def percentage(from_,  to_):
        if from_ is None or from_==0:
            return None
        if to_ is None or to_==0:
            return None
        return (to_-from_)/from_
    ###########################
    
    first_value_not_noz=ymv_transposition_first_value_not_noz(lod_ymv_transposition)
    r=[]
    if first_value_not_noz is None:
        return r
    for i, d in enumerate(lod_ymv_transposition):
        new_d={"year":d["year"]}
        new_d["m1"]=None if i==0 else percentage(lod_ymv_transposition[i-1]["m12"], d["m1"])
        new_d["m2"]=percentage(d["m1"], d["m2"])
        new_d["m3"]=percentage(d["m2"], d["m3"])
        new_d["m4"]=percentage(d["m3"], d["m4"])
        new_d["m5"]=percentage(d["m4"], d["m5"])
        new_d["m6"]=percentage(d["m5"], d["m6"])
        new_d["m7"]=percentage(d["m6"], d["m7"])
        new_d["m8"]=percentage(d["m7"], d["m8"])
        new_d["m9"]=percentage(d["m8"], d["m9"])
        new_d["m10"]=percentage(d["m9"], d["m10"])
        new_d["m11"]=percentage(d["m10"], d["m11"])
        new_d["m12"]=percentage(d["m11"], d["m12"])
        key_first_value=d_ymv_transposition_first_key_not_noz(d)
        key_last_value=d_ymv_transposition_last_key_not_noz(d)
        if i==0:
            new_d["total"]=None if key_first_value is None or key_last_value is None else percentage(d[key_first_value], d[key_last_value])
        else:
            key_previous_last_value=d_ymv_transposition_last_key_not_noz(lod_ymv_transposition[i-1])
            new_d["total"]=None if key_previous_last_value is None or key_last_value is None else percentage(lod_ymv_transposition[i-1][key_previous_last_value], d[key_last_value])
        new_d["from_first_quote"]=None if key_last_value is None else percentage(first_value_not_noz["value"], d[key_last_value])
        r.append(new_d)
    return r

def lod_ymv_transposition_sum(lymv_a, lymv_b):
    """
        Sums to lod_ymv_transpositions
    """
    def get_younger(year, field):
        if year in d_younger:
            return d_younger[year][field]
        else:
            return 0
    
    if len(lymv_a)==0:
        return lymv_b
    if len(lymv_b)==0:
        return lymv_a
    year_lymv_a=lymv_a[0]["year"]
    year_lymv_b=lymv_b[0]["year"]
    print(year_lymv_a, year_lymv_b)
    older=lymv_a if year_lymv_a<year_lymv_b else lymv_b
    younger=lymv_a if year_lymv_a>year_lymv_b else lymv_b
    d_younger=lod.lod2dod(younger, "year")
    r=[]
    for d in older:
        new={}
        new["year"]=d["year"]
        new["m1"]=d["m1"]+get_younger(d["year"],"m1")
        new["m2"]=d["m2"]+get_younger(d["year"],"m2")
        new["m3"]=d["m3"]+get_younger(d["year"],"m3")
        new["m4"]=d["m4"]+get_younger(d["year"],"m4")
        new["m5"]=d["m5"]+get_younger(d["year"],"m5")
        new["m6"]=d["m6"]+get_younger(d["year"],"m6")
        new["m7"]=d["m7"]+get_younger(d["year"],"m7")
        new["m8"]=d["m8"]+get_younger(d["year"],"m8")
        new["m9"]=d["m9"]+get_younger(d["year"],"m9")
        new["m10"]=d["m10"]+get_younger(d["year"],"m10")
        new["m11"]=d["m11"]+get_younger(d["year"],"m11")
        new["m12"]=d["m12"]+get_younger(d["year"],"m12")
        new["total"]=d["total"]+get_younger(d["year"],"total")
        r.append(new)
    return r

## Converts a tipical groyp by lor with year, month (normaly extracted from db to fill empty values)
def lod_ymv_filling(lod, year_from, year_to=date.today().year, fill_value=0, key_year="year", key_month="month", key_value="value"):
    ld_tuple=lod.lod2dod_tuple(lod, key_year, key_month)
    for year in range(year_from,year_to+1):
        for month in range (1,13):
            if not (key_year,key_month) in ld_tuple:
                ld_tuple[(key_year,key_month)]={key_year: year, key_month: month, key_value:fill_value}
    r=[]
    for d in ld_tuple.values:
        r.append(d)
    return r
